//only mode 0 is implemented


struct ppi8255struct {
	
	unsigned char dir[3];//port direction, 1 is output, 0 is input
	unsigned char reg[4];//regs 0..2 is port value, 3 is control word
	
	int mode;
	
};

enum {
	PPI8255_PORTA=0,
	PPI8255_PORTB,
	PPI8255_PORTC,
	PPI8255_CWORD
};

enum {
	PPI8255_MODE0=0,
	PPI8255_MODE1,
	PPI8255_MODE2
};

#define PPI8255_PIN_D0	0x01
#define PPI8255_PIN_D1	0x02
#define PPI8255_PIN_D2	0x04
#define PPI8255_PIN_D3	0x08
#define PPI8255_PIN_D4	0x10
#define PPI8255_PIN_D5	0x20
#define PPI8255_PIN_D6	0x40
#define PPI8255_PIN_D7	0x80



inline void ppi_set_pin(ppi8255struct *p,int pnum,int pin,bool state)
{
	if(state) p->reg[pnum]|=pin; else p->reg[pnum]&=~pin;
}



inline bool ppi_get_pin(ppi8255struct *p,int pnum,int pin)
{
	return (p->reg[pnum]&pin)?true:false;
}



inline void ppi_set_pins(ppi8255struct *p,int pnum,unsigned char state)
{
	p->reg[pnum]=state;
}



inline unsigned char ppi_get_pins(ppi8255struct *p,int pnum)
{
	return p->reg[pnum];
}



inline unsigned char ppi_reg_rd(ppi8255struct *p,unsigned char adr)
{
	adr&=3;

	return p->reg[adr];
}



inline void ppi_reg_wr(ppi8255struct *p,unsigned char adr,unsigned char val)
{
	adr&=3;
	
	switch(adr)
	{
	case PPI8255_CWORD:

		p->reg[PPI8255_CWORD]=val;

		if(val&0x80)	//mode select
		{
			switch((val>>5)&3)
			{
				case 0:
				{
					p->mode=PPI8255_MODE0;
					
					p->dir[PPI8255_PORTA]= (val&0x10)?0x00:0xff;
					p->dir[PPI8255_PORTB]= (val&0x02)?0x00:0xff;
					p->dir[PPI8255_PORTC]=((val&0x08)?0x00:0xf0)|((val&0x01)?0x00:0x0f);
				}
				break;

				case 1:
				{
					p->mode=PPI8255_MODE1;
					
					log_add("ppi unsupported mode 1 requested\n");
				}
				break;

				default:
				{
					p->mode=PPI8255_MODE2;
					
					log_add("ppi unsupported mode 2 requested\n");
				}
			}
		}
		else	//channel C bits set/reset
		{
			if(val&0x01)	//set bit
			{
				p->reg[PPI8255_PORTC]|=((1<<((val&0x0e)>>1))&p->dir[PPI8255_PORTC]);
			}
			else			//reset bit
			{
				p->reg[PPI8255_PORTC]&=~((1<<((val&0x0e)>>1))&p->dir[PPI8255_PORTC]);
			}
		}
		break;
		
	case PPI8255_PORTA: p->reg[PPI8255_PORTA]=val; break;
	case PPI8255_PORTB: p->reg[PPI8255_PORTB]=val; break;
	case PPI8255_PORTC: p->reg[PPI8255_PORTC]=val; break;
	}
}



void ppi_reset(ppi8255struct *p)
{
	memset(p,0,sizeof(ppi8255struct));

	ppi_reg_wr(p,PPI8255_CWORD,0x9b);	//set default mode
}