//---------------------------------------------------------------------------

#include <vcl.h>
#include <stdio.h>

#pragma hdrstop

#include "UnitMain.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TFormMain *FormMain;



#include "log.h"
#include "config.h"
#include "waveout.h"

#include "zymosis.h"
//#include "8253.h"
#include "8255.h"
#include "ChipPIT8253.h"


#define WINDOW_TITLE		"AONZ80Emu v1.01"

#define SAMPLE_RATE			44100

#define MASTER_CLOCK		4000000

#define DISPLAY_TIMEOUT		1800

#define TURBO_FACTOR		15

#define BUTTON_WDT			67
#define BUTTON_HGT			40
#define BUTTON_SPACING		12
#define BUTTON_EXT_SPACING 	20
#define BUTTON_CURVE		24

int rom_size=2048;
int ram_size=2048;

struct {

	unsigned char rom[65536];
	unsigned char ram[8192];

	Z80Info z80;
	ppi8255struct port;

	ChipPIT8253 timer;

	unsigned char segment;
	unsigned char select;
	unsigned char diodes;

	bool led1;
	bool led2;

	int cpu_ticks;

	bool clock_phase;

	int sound_level;

	bool handset_down;

} AON;

int debugRAMOffset;

bool keyState[32];

int displayState[9][8];

enum {
	key1=0,
	key2,
	key3,
	key4,
	key5,
	key6,
	key7,
	key8,
	key9,
	keyAsterisk,
	key0,
	keyNumber,
	keyB0,
	keyAlarm,
	keyPause,
	keyStore,
	keyRedial,
	keyB1,
	keySound,
	keyDate,
	keyReset,
	keyFlash,
	keyB2,
	keySec,
	keySay,
	keySet,
	keyCorr,
	keyMode,
	keyA0,
	keyA1,
	keyA2,
	keyA3
};

const char *keyNames[]={"1","2","3","4","5","6","7","8","9","*","0","#","B0","ALARM","PAUSE","STORE","REDIAL","B1","SOUND","DATE","RESET","FLASH","B2","SEC","SAY","SET","CORR","MODE","A0","A1","A2","A3"};


AnsiString romName;
AnsiString configName;

int soundBufferSize;
int soundBufferCount;



void reset_keyboard(void)
{
	memset(keyState,0,sizeof(keyState));
}



unsigned char keyboard_poll_row(void)
{
	unsigned char row;

	row=0;

	//diodes

	if(AON.select==0)
	{
		row=AON.diodes&0xf0;

		return row;
	}

	if(!(AON.diodes&0x80))	//standard keyboard
	{
		switch(AON.select)
		{
		case 1:
			{
				row|=keyState[key7]?0x80:0x00;
				row|=keyState[key1]?0x40:0x00;
			}
			break;

		case 2:
			{
				row|=keyState[key8]?0x80:0x00;
				row|=keyState[key2]?0x40:0x00;
			}
			break;

		case 3:
			{
				row|=keyState[key9]?0x80:0x00;
				row|=keyState[key3]?0x40:0x00;
			}
			break;

		case 4:
			{
				row|=keyState[keyAsterisk]?0x80:0x00;
				row|=keyState[key4       ]?0x40:0x00;
			}
			break;

		case 5:
			{
				row|=keyState[key0]?0x80:0x00;
				row|=keyState[key5]?0x40:0x00;
			}
			break;

		case 6:
			{
				row|=keyState[keyNumber]?0x80:0x00;
				row|=keyState[key6]?0x40:0x00;
			}
			break;

		}
	}
	else
	{	//extended keyboard

		switch(AON.select)
		{
		case 1:
			{
				row|=keyState[key1 ]?0x80:0x00;
				row|=keyState[key2 ]?0x40:0x00;
				row|=keyState[key3 ]?0x20:0x00;
				row|=keyState[keyB0]?0x10:0x00;
			}
			break;

		case 2:
			{
				row|=keyState[key4 ]?0x80:0x00;
				row|=keyState[key5 ]?0x40:0x00;
				row|=keyState[key6 ]?0x20:0x00;
				row|=keyState[keyB1]?0x10:0x00;
			}
			break;

		case 3:
			{
				row|=keyState[key7 ]?0x80:0x00;
				row|=keyState[key8 ]?0x40:0x00;
				row|=keyState[key9 ]?0x20:0x00;
				row|=keyState[keyB2]?0x10:0x00;
			}
			break;

		case 4:
			{
				row|=keyState[keyAsterisk]?0x80:0x00;
				row|=keyState[key0       ]?0x40:0x00;
				row|=keyState[keyNumber  ]?0x20:0x00;
				row|=keyState[keyMode    ]?0x10:0x00;
			}
			break;

		case 5:
			{
				row|=keyState[keyRedial]?0x80:0x00;
				row|=keyState[keyStore ]?0x40:0x00;
				row|=keyState[keyPause ]?0x20:0x00;
				row|=keyState[keyAlarm ]?0x10:0x00;
			}
			break;

		case 6:
			{
				row|=keyState[keyFlash]?0x80:0x00;
				row|=keyState[keyReset]?0x40:0x00;
				row|=keyState[keyDate ]?0x20:0x00;
				row|=keyState[keySound]?0x10:0x00;
			}
			break;

		case 7:
			{
				row|=keyState[keyCorr]?0x80:0x00;
				row|=keyState[keySet ]?0x40:0x00;
				row|=keyState[keySay ]?0x20:0x00;
				row|=keyState[keySec ]?0x10:0x00;
			}
			break;

		case 8:
			{
				row|=keyState[keyA0]?0x80:0x00;
				row|=keyState[keyA1]?0x40:0x00;
				row|=keyState[keyA2]?0x20:0x00;
				row|=keyState[keyA3]?0x10:0x00;
			}
			break;
		}
	}

	return row;
}



void display_update(void)
{
	if(AON.select<9)
	{
		if(AON.segment&0x01) displayState[AON.select][0]=DISPLAY_TIMEOUT;
		if(AON.segment&0x02) displayState[AON.select][1]=DISPLAY_TIMEOUT;
		if(AON.segment&0x04) displayState[AON.select][2]=DISPLAY_TIMEOUT;
		if(AON.segment&0x08) displayState[AON.select][3]=DISPLAY_TIMEOUT;
		if(AON.segment&0x10) displayState[AON.select][4]=DISPLAY_TIMEOUT;
		if(AON.segment&0x20) displayState[AON.select][5]=DISPLAY_TIMEOUT;
		if(AON.segment&0x40) displayState[AON.select][6]=DISPLAY_TIMEOUT;
		if(AON.segment&0x80) displayState[AON.select][7]=DISPLAY_TIMEOUT;
	}
}



static void z80MemWrite(Z80Info *z80,uint16_t addr,uint8_t value,Z80MemIOType mio)
{
	if(addr&0x2000) AON.ram[addr&(ram_size-1)]=value;
}



static uint8_t z80MemRead(Z80Info *z80,uint16_t addr,Z80MemIOType mio)
{
	if(addr&0x2000)
	{
		return AON.ram[addr&(ram_size-1)];
	}
	else
	{
		addr=(addr&0x1fff)|((addr>>1)&0xe000);	//translate ROM address lines

		if(ppi_get_pin(&AON.port,PPI8255_PORTA,PPI8255_PIN_D6)) addr|=0x8000;	//second 32K bank

		return AON.rom[addr&(rom_size-1)];
	}
}



static uint8_t z80PortIn(Z80Info *z80,uint16_t adr,Z80PIOType pio)
{
	unsigned char row,reg;

	reg=((adr>>3)&3);

	if(!(adr&0x20))//8255
	{
		switch(reg)
		{
		case 0: return ppi_reg_rd(&AON.port,PPI8255_PORTA);

		case 1:
			{
				row=keyboard_poll_row();

				ppi_set_pin(&AON.port,PPI8255_PORTB,PPI8255_PIN_D4,(row^0xf0)&0x10?true:false);
				ppi_set_pin(&AON.port,PPI8255_PORTB,PPI8255_PIN_D5,(row^0xf0)&0x20?true:false);
				ppi_set_pin(&AON.port,PPI8255_PORTB,PPI8255_PIN_D6,(row^0xf0)&0x40?true:false);
				ppi_set_pin(&AON.port,PPI8255_PORTB,PPI8255_PIN_D7,(row^0xf0)&0x80?true:false);

				return ppi_reg_rd(&AON.port,PPI8255_PORTB);
			}

		case 2: return ppi_reg_rd(&AON.port,PPI8255_PORTC);
		case 3: return ppi_reg_rd(&AON.port,PPI8255_CWORD);
		}
	}

	if(!(adr&0x40))//8253
	{
		switch(reg)
		{
		case 0: return AON.timer.CpuRead(CT_0);
		case 1: return AON.timer.CpuRead(CT_1);
		case 2: return AON.timer.CpuRead(CT_2);
		case 3: return AON.timer.CpuRead(CT_CWR);
		}
	}

	return 0xff;//pull up to +5v, some software won't work if it is missing
}


static void z80PortOut(Z80Info *z80,uint16_t adr,uint8_t value,Z80PIOType pio)
{
	unsigned char reg;

	reg=((adr>>3)&3);

	if(!(adr&0x20))//8255
	{
		switch(reg)
		{
		case 0:
			{
				ppi_reg_wr(&AON.port,PPI8255_PORTA,value);

				if(ppi_get_pin(&AON.port,PPI8255_PORTA,PPI8255_PIN_D3))
				{
					AON.select=0;
				}
				else
				{
					AON.select=(ppi_get_pins(&AON.port,PPI8255_PORTA)&7)+1;
				}
			}
			break;

		case 1:
			{
				ppi_reg_wr(&AON.port,PPI8255_PORTB,value);
			}
			break;

		case 2:
			{
				ppi_reg_wr(&AON.port,PPI8255_PORTC,value);

				AON.led1=ppi_get_pin(&AON.port,PPI8255_PORTC,PPI8255_PIN_D0)?false:true;
				AON.led2=ppi_get_pin(&AON.port,PPI8255_PORTC,PPI8255_PIN_D3)?false:true;
			}
			break;

		case 3:
			{
				ppi_reg_wr(&AON.port,PPI8255_CWORD,value);
			}
			break;
		}
	}

	if(!(adr&0x40))//8253
	{
		switch(reg)
		{
		case 0: AON.timer.CpuWrite(CT_0,value); break;
		case 1: AON.timer.CpuWrite(CT_1,value); break;
		case 2: AON.timer.CpuWrite(CT_2,value); break;
		case 3: AON.timer.CpuWrite(CT_CWR,value); break;
		}
	}

	if(!(adr&0x80)) //display
	{
		AON.segment=value;
	}

	display_update();
}



void aon_reset(void)
{
	Z80_Reset(&AON.z80);

	ppi_reset(&AON.port);

	AON.segment=0;
	AON.select=0;
	AON.handset_down=true;

	reset_keyboard();
}



void aon_clear_display(void)
{
	int i,j;

	for(i=0;i<9;++i) for(j=0;j<8;++j) displayState[i][j]=0;
}



void aon_fade_display(void)
{
	int i,j;

	for(i=0;i<9;++i) for(j=0;j<8;++j) if(displayState[i][j]) --displayState[i][j];
}



void aon_init(void)
{
	int i;

	Z80_ResetCallbacks(&AON.z80);

	AON.z80.memReadFn =z80MemRead;
	AON.z80.memWriteFn=z80MemWrite;
	AON.z80.portInFn  =z80PortIn;
	AON.z80.portOutFn =z80PortOut;
	AON.z80.evenM1    =false;

	aon_reset();

	for(i=0;i<8192;++i) AON.ram[i]=rand()&255;

	aon_clear_display();
}



bool aon_load_rom(const char *filename)
{
	FILE *file;
	int size;

	file=fopen(filename,"rb");

	if(!file) return false;

	fseek(file,0,SEEK_END);

	size=ftell(file);

	fseek(file,0,SEEK_SET);

	if(size!=2048&&size!=8192&&size!=16384&&size!=32768&&size!=65536)
	{
		fclose(file);
		return false;
	}

	fread(AON.rom,size,1,file);
	fclose(file);

	rom_size=size;

	aon_reset();

	return true;
}



void aon_update(int ticks)
{
	int i,t,clock;
	bool ir,prev;
	float sample,mul;

	if(!ticks) return;

	sample=0;
	mul=1.0f/(float)ticks/2.0f;

	ppi_set_pin(&AON.port,PPI8255_PORTB,PPI8255_PIN_D2,AON.handset_down);

	while(ticks>0)
	{
		t=Z80_ExecuteStep(&AON.z80);

		AON.cpu_ticks+=t;
		ticks-=t;

		if(AON.cpu_ticks>=2)
		{
			AON.timer.PeripheralSetGate(CT_0,true);
			AON.timer.PeripheralSetGate(CT_1,ppi_get_pin(&AON.port,PPI8255_PORTA,PPI8255_PIN_D6));
			AON.timer.PeripheralSetGate(CT_2,ppi_get_pin(&AON.port,PPI8255_PORTA,PPI8255_PIN_D7));

			ir=false;

			while(AON.cpu_ticks>=2)
			{
				prev=AON.timer.PeripheralReadOut(CT_0);

				AON.timer.PeripheralSetClock(CT_0,AON.clock_phase);
				AON.timer.PeripheralSetClock(CT_1,AON.clock_phase);
				AON.timer.PeripheralSetClock(CT_2,AON.clock_phase);

			   	AON.clock_phase^=true;

				if(prev!=AON.timer.PeripheralReadOut(CT_0)) ir=true;

				if(ppi_get_pin(&AON.port,PPI8255_PORTA,PPI8255_PIN_D4))
				{
					if(AON.timer.PeripheralReadOut(CT_2)) sample+=mul;
				}

				//if(ppi_get_pin(&AON.port,PPI8255_PORTC,PPI8255_PIN_D0)) sample+=mul;	//out dial
				//if(ppi_get_pin(&AON.port,PPI8255_PORTC,PPI8255_PIN_D2)) sample+=mul;  //sound&voice

				AON.cpu_ticks-=2;
			}

			if(ir) AON.cpu_ticks+=Z80_Interrupt(&AON.z80);
		}
	}

	display_update();

	AON.sound_level=(int)(8192.0f*sample);
}



void sound_render(short int *wave,int size)
{
	int i,smp,frames;
	bool run;

	run=false;

	if(FormMain->SpeedButtonRun->Down)
	{
		if(FormMain->SpeedButtonTurbo->Down) frames=TURBO_FACTOR; else frames=1;

		run=true;
	}

	if(FormMain->SpeedButtonAdvance->Down)
	{
		FormMain->SpeedButtonAdvance->Down=false;

		frames=1;
		run=true;
	}

	if(!run) return;

	for(i=0;i<size/2;++i)
	{
		aon_fade_display();

		aon_update(MASTER_CLOCK*frames/SAMPLE_RATE-1);

		smp=AON.sound_level;

		*wave++=smp;
		*wave++=smp;
	}

	if(run) FormMain->UpdateEmu();
}



void __fastcall TFormMain::DrawKey(TCanvas *c,int x,int y,AnsiString label,bool state)
{
	c->Pen->Width=3;
	c->Pen->Color=clBlack;
	c->Brush->Color=state?clSilver:clWhite;
	c->RoundRect(x,y,x+BUTTON_WDT,y+BUTTON_HGT,BUTTON_CURVE,BUTTON_CURVE);
	c->Font->Color=clBlack;
	c->Font->Size=11;
	c->TextOut(x+BUTTON_WDT/2-c->TextWidth(label)/2,y+BUTTON_HGT/2-c->TextHeight(label)/2,label);
}



void __fastcall TFormMain::DrawKeyboard(void)
{
	int i,j,sx,sy,button;
	TCanvas *c;

	if(!GroupBoxKeypad->Visible) return;

	c=ImageKeypad->Canvas;

	c->Lock();

	c->Brush->Color=clBtnFace;
	c->FillRect(TRect(0,0,ImageKeypad->Width,ImageKeypad->Height));

	button=0;
	sy=1;

	for(i=0;i<4;++i)
	{
		sx=1;

		for(j=0;j<3;++j)
		{
			DrawKey(c,sx,sy,keyNames[button],keyState[button]);

			sx+=BUTTON_WDT+BUTTON_SPACING;

			++button;
		}

		sy+=BUTTON_HGT+BUTTON_SPACING;
	}

	if(AON.diodes&0x80)
	{
		sy=0;

		for(i=0;i<4;++i)
		{
			sx=(BUTTON_WDT+BUTTON_SPACING)*3+BUTTON_EXT_SPACING;

			for(j=0;j<5;++j)
			{
				DrawKey(c,sx,sy,keyNames[button],keyState[button]);

				sx+=BUTTON_WDT+BUTTON_SPACING;

				++button;
			}

			sy+=BUTTON_HGT+BUTTON_SPACING;
		}
	}

	c->Unlock();
}



void __fastcall TFormMain::CheckKeyboard(int x,int y,bool down)
{
	int button;

	button=-1;

	if(y>=0&&y<(BUTTON_HGT+BUTTON_SPACING)*4)
	{
		if(x>=0&&x<(BUTTON_WDT+BUTTON_SPACING)*3)
		{
			if(x%(BUTTON_WDT+BUTTON_SPACING)<BUTTON_WDT&&y%(BUTTON_HGT+BUTTON_SPACING)<BUTTON_HGT)
			{
				button=(y/(BUTTON_HGT+BUTTON_SPACING))*3+(x/(BUTTON_WDT+BUTTON_SPACING));
			}
		}

		if(x>=(BUTTON_WDT+BUTTON_SPACING)*3+BUTTON_EXT_SPACING&&x<(BUTTON_WDT+BUTTON_SPACING)*3+BUTTON_EXT_SPACING+(BUTTON_WDT+BUTTON_SPACING)*5)
		{
			x-=(BUTTON_WDT+BUTTON_SPACING)*3+BUTTON_EXT_SPACING;

			if(x%(BUTTON_WDT+BUTTON_SPACING)<BUTTON_WDT&&y%(BUTTON_HGT+BUTTON_SPACING)<BUTTON_HGT)
			{
				button=12+(y/(BUTTON_HGT+BUTTON_SPACING))*5+(x/(BUTTON_WDT+BUTTON_SPACING));
			}
		}
	}

	if(button>=0)
	{
		if(down)
		{
			if(!(GetKeyState(VK_CONTROL)&0x8000)) reset_keyboard();
		}

		keyState[button]=down;

		DrawKeyboard();
	}
}



void __fastcall TFormMain::ReleaseKeys(void)
{
	reset_keyboard();
}



void __fastcall TFormMain::DrawLEDs(TCanvas *c,int x,int y)
{
	ShapeLED1->Brush->Color=(TColor)(AON.led1?0x2040ff:0x000040);
	ShapeLED1->Pen  ->Color=(TColor)(AON.led1?0x000020:0x000000);

	ShapeLED2->Brush->Color=(TColor)(AON.led2?0x2040ff:0x000040);
	ShapeLED2->Pen  ->Color=(TColor)(AON.led2?0x000020:0x000000);
}



void __fastcall TFormMain::DrawDisplay(TCanvas *c,int x,int y)
{
	int i,j;

	c->Lock();

	for(i=0;i<9;++i)
	{
		ImageListSegments->Draw(c,x,y,8);

		for(j=0;j<8;++j)
		{
			if(displayState[i][j]) ImageListSegments->Draw(c,x,y,j);
		}

		x+=68;
	}

	c->Unlock();
}



void __fastcall TFormMain::DrawDebugInfo(TImage *img)
{
	int i,j,x,y,h,ptr;
	AnsiString str;
	TCanvas *c;

	if(!GroupBoxDebug->Visible) return;

	c=img->Canvas;

	c->Lock();

	//z80

	y=0;
	h=18;

	c->TextOut(0,y,"Z80 regs"); y+=22;

	c->TextOut(0,y+0*h,"PC :"+IntToHex(AON.z80.pc,4));
	c->TextOut(0,y+1*h,"SP :"+IntToHex(AON.z80.sp.w,4));
	c->TextOut(0,y+2*h,"AF :"+IntToHex(AON.z80.af.w,4));
	c->TextOut(0,y+3*h,"BC :"+IntToHex(AON.z80.bc.w,4));
	c->TextOut(0,y+4*h,"DE :"+IntToHex(AON.z80.de.w,4));
	c->TextOut(0,y+5*h,"HL :"+IntToHex(AON.z80.hl.w,4));
	c->TextOut(0,y+6*h,"AF':"+IntToHex(AON.z80.afx.w,4));
	c->TextOut(0,y+7*h,"BC':"+IntToHex(AON.z80.bcx.w,4));
	c->TextOut(0,y+8*h,"DE':"+IntToHex(AON.z80.dex.w,4));
	c->TextOut(0,y+9*h,"HL':"+IntToHex(AON.z80.hlx.w,4));
	c->TextOut(0,y+10*h,"IX :"+IntToHex(AON.z80.ix.w,4));
	c->TextOut(0,y+11*h,"IY :"+IntToHex(AON.z80.iy.w,4));

	//8255

	x=90;
	y=0;

	c->TextOut(x,y,"8255 out/dir"); y+=22;

	c->TextOut(x,y+0*h,"PortA: "+IntToHex(AON.port.reg[0],2)+" "+IntToHex(AON.port.dir[0],2));
	c->TextOut(x,y+1*h,"PortB: "+IntToHex(AON.port.reg[1],2)+" "+IntToHex(AON.port.dir[1],2));
	c->TextOut(x,y+2*h,"PortC: "+IntToHex(AON.port.reg[2],2)+" "+IntToHex(AON.port.dir[2],2));
	c->TextOut(x,y+3*h,"CWord: "+IntToHex(AON.port.reg[3],2));

	//8253

	x=90;
	y=100;

	c->TextOut(x,y,"8253 reg"); y+=22;

	c->TextOut(x,y+0*h,"RegA : "+IntToHex(AON.timer.Counters[0].InitValue,4));
	c->TextOut(x,y+1*h,"RegB : "+IntToHex(AON.timer.Counters[1].InitValue,4));
	c->TextOut(x,y+2*h,"RegC : "+IntToHex(AON.timer.Counters[2].InitValue,4));
	c->TextOut(x,y+3*h,"CWord: "+IntToHex(AON.timer.Counters[0].CWR,2));

	//HW regs

	x=90;
	y=200;

	c->TextOut(x,y+0*h,"Segment: "+IntToHex(AON.segment,2));
	c->TextOut(x,y+1*h,"Select :  "+IntToHex(AON.select,1));

	//RAM

	ptr=debugRAMOffset&65535;

	y=0;
	h=18;

	for(i=0;i<2048/16;++i)
	{
		x=210;

		c->TextOut(x,y,IntToHex(ptr,4)+":");

		x+=44;

		for(j=0;j<16;++j)
		{
			c->TextOut(x,y,IntToHex(z80MemRead(NULL,ptr,(Z80MemIOType)0),2));

			ptr=(ptr+1)&65535;

			x+=18;
		}

		y+=h;

		if(y+h>=img->Height) break;
	}

	c->Unlock();
}



void __fastcall TFormMain::UpdateWindowsView(void)
{
	int sy;

	GroupBoxKeypad->Visible=MViewKeypad->Checked;
	GroupBoxDebug ->Visible=MViewDebug->Checked;

	sy=GroupBoxDisplay->Top+GroupBoxDisplay->Height;

	if(GroupBoxKeypad->Visible)
	{
		GroupBoxKeypad->Top=sy;
		sy+=GroupBoxKeypad->Height;
	}

	if(GroupBoxDebug->Visible)
	{
		GroupBoxDebug->Top=sy;
		sy+=GroupBoxDebug->Height;
	}

	ClientHeight=sy+4;

	Caption=AnsiString(WINDOW_TITLE)+" ["+romName+"]";
}



void __fastcall TFormMain::ConfigLoad(void)
{
	FILE *file;

	//try load config, use default settings if not available

	config_open(configName.c_str());

	romName=config_read_string("ROM","");

	MViewKeypad->Checked=config_read_int("ViewKeypad",1)?true:false;
	MViewDebug ->Checked=config_read_int("ViewDebug" ,1)?true:false;

	soundBufferSize=config_read_int("SoundBufferSize",4096);
	soundBufferCount=config_read_int("SoundBufferCount",8);

	MNoDiodes->Checked=config_read_int("NoDiodes",1)?true:false;
	MDiodesVD1->Checked=config_read_int("DiodesVD1",0)?true:false;
	MDiodesVD2->Checked=config_read_int("DiodesVD2",0)?true:false;
	MDiodesVD3->Checked=config_read_int("DiodesVD3",0)?true:false;
	MDiodesVD4->Checked=config_read_int("DiodesVD4",0)?true:false;

	config_close();
}



void __fastcall TFormMain::ConfigSave(void)
{
	FILE *file;

	file=fopen(configName.c_str(),"wt");

	fprintf(file,"ROM=%s\n",romName.c_str());
	fprintf(file,"ViewKeypad=%i\n",MViewKeypad->Checked?1:0);
	fprintf(file,"ViewDebug=%i\n",MViewDebug->Checked?1:0);

	fprintf(file,"SoundBufferSize=%i\n",soundBufferSize);
	fprintf(file,"SoundBufferCount=%i\n",soundBufferCount);

	fprintf(file,"NoDiodes=%i\n",MNoDiodes->Checked?1:0);
	fprintf(file,"DiodesVD1=%i\n",MDiodesVD1->Checked?1:0);
	fprintf(file,"DiodesVD2=%i\n",MDiodesVD2->Checked?1:0);
	fprintf(file,"DiodesVD3=%i\n",MDiodesVD3->Checked?1:0);
	fprintf(file,"DiodesVD4=%i\n",MDiodesVD4->Checked?1:0);

	fclose(file);
}



void __fastcall TFormMain::LoadROM(void)
{
	if(romName=="") return;

	if(!aon_load_rom(romName.c_str()))
	{
		Application->MessageBox("Can't load ROM file","Error",MB_OK);
	}
	else
	{
		aon_init();

		OpenDialogROM->FileName=romName;
	}
}



void __fastcall TFormMain::UpdateEmu(void)
{
	UpdateActive=true;

	DrawDisplay(ImageDisplay->Canvas,27,35);
	DrawLEDs(ImageDisplay->Canvas,0,0);

	UpdateActive=false;
}



//---------------------------------------------------------------------------
__fastcall TFormMain::TFormMain(TComponent* Owner)
: TForm(Owner)
{
}
//---------------------------------------------------------------------------

void __fastcall TFormMain::FormCreate(TObject *Sender)
{
	romName="";
	configName=ParamStr(0).SubString(0,ParamStr(0).LastDelimiter("."))+"cfg";

	debugRAMOffset=65536-8192;

	ImageDisplay->Picture=new TPicture();
	ImageDisplay->Picture->Bitmap=new Graphics::TBitmap();
	ImageDisplay->Picture->Bitmap->PixelFormat=pf24bit;
	ImageDisplay->Picture->Bitmap->SetSize(ImageDisplay->Width,ImageDisplay->Height);

	ImageDisplay->Canvas->Brush->Color=clBlack;
	ImageDisplay->Canvas->FillRect(TRect(0,0,ImageDisplay->Picture->Bitmap->Width,ImageDisplay->Picture->Bitmap->Height));

	ImageKeypad->Picture=new TPicture();
	ImageKeypad->Picture->Bitmap=new Graphics::TBitmap();
	ImageKeypad->Picture->Bitmap->PixelFormat=pf24bit;
	ImageKeypad->Picture->Bitmap->SetSize(ImageKeypad->Width,ImageKeypad->Height);

	ImageKeypad->Canvas->Brush->Color=clBtnFace;
	ImageKeypad->Canvas->FillRect(TRect(0,0,ImageKeypad->Picture->Bitmap->Width,ImageKeypad->Picture->Bitmap->Height));

	ImageDebug->Picture=new TPicture();
	ImageDebug->Picture->Bitmap=new Graphics::TBitmap();
	ImageDebug->Picture->Bitmap->PixelFormat=pf24bit;
	ImageDebug->Picture->Bitmap->SetSize(ImageDebug->Width,ImageDebug->Height);

	ImageDebug->Canvas->Brush->Color=clBtnFace;
	ImageDebug->Canvas->FillRect(TRect(0,0,ImageDebug->Picture->Bitmap->Width,ImageDebug->Picture->Bitmap->Height));

	ImageDebug->Canvas->Font->Name="Fixedsys";

	//log_open("log.txt");

	DoubleBuffered=true;

	ConfigLoad();

	if(ParamStr(1)!="") romName=ParamStr(1);

	MNoDiodesClick(Sender);

	aon_init();

	LoadROM();

	UpdateWindowsView();

	UpdateActive=false;
	//sampleCount=0;

	waveout_init(Handle,SAMPLE_RATE,soundBufferSize,soundBufferCount);
}
//---------------------------------------------------------------------------

void __fastcall TFormMain::FormMouseLeave(TObject *Sender)
{
	ReleaseKeys();
}
//---------------------------------------------------------------------------

void __fastcall TFormMain::FormMouseMove(TObject *Sender, TShiftState Shift,
int X, int Y)
{
	CheckKeyboard(X,Y,Shift.Contains(ssLeft));
}
//---------------------------------------------------------------------------

void __fastcall TFormMain::FormDestroy(TObject *Sender)
{
	waveout_shut();

	ConfigSave();

	log_close();
}
//---------------------------------------------------------------------------

void __fastcall TFormMain::MResetClick(TObject *Sender)
{
	SpeedButtonResetClick(Sender);
}
//---------------------------------------------------------------------------


void __fastcall TFormMain::ImageKeypadMouseDown(TObject *Sender,
TMouseButton Button, TShiftState Shift, int X, int Y)
{
	CheckKeyboard(X,Y,Shift.Contains(ssLeft));
}
//---------------------------------------------------------------------------

void __fastcall TFormMain::ImageKeypadMouseMove(TObject *Sender,
TShiftState Shift, int X, int Y)
{
	CheckKeyboard(X,Y,Shift.Contains(ssLeft));
}
//---------------------------------------------------------------------------

void __fastcall TFormMain::ImageKeypadMouseUp(TObject *Sender,
TMouseButton Button, TShiftState Shift, int X, int Y)
{
	ReleaseKeys();	
}
//---------------------------------------------------------------------------

void __fastcall TFormMain::FormPaint(TObject *Sender)
{
	DrawKeyboard();
}
//---------------------------------------------------------------------------

void __fastcall TFormMain::FormKeyDown(TObject *Sender, WORD &Key,
TShiftState Shift)
{
	int button;

	button=-1;

	switch(Key)
	{
	case VK_NEXT:  debugRAMOffset+=128; break;
	case VK_PRIOR: debugRAMOffset-=128; break;

	case '0': case VK_NUMPAD0: button=key0; break;
	case '1': case VK_NUMPAD1: button=key1; break;
	case '2': case VK_NUMPAD2: button=key2; break;
	case '3': case VK_NUMPAD3: button=key3; break;
	case '4': case VK_NUMPAD4: button=key4; break;
	case '5': case VK_NUMPAD5: button=key5; break;
	case '6': case VK_NUMPAD6: button=key6; break;
	case '7': case VK_NUMPAD7: button=key7; break;
	case '8': case VK_NUMPAD8: button=key8; break;
	case '9': case VK_NUMPAD9: button=key9; break;

	case VK_ESCAPE:
	case VK_DIVIDE:
	case VK_ADD:
		button=keyNumber;
		break;

	case VK_RETURN:
	case VK_MULTIPLY:
		button=keyAsterisk;
		break;
	}

	if(button>=0)
	{
		if(!(GetKeyState(VK_CONTROL)&0x8000)) reset_keyboard();

		keyState[button]=TRUE;
		DrawKeyboard();
	}
}
//---------------------------------------------------------------------------

void __fastcall TFormMain::FormKeyUp(TObject *Sender, WORD &Key,
TShiftState Shift)
{
	if((Key>='0'&&Key<='9')||(Key>=VK_NUMPAD0&&Key<=VK_NUMPAD9)||Key==VK_ESCAPE||Key==VK_RETURN||Key==VK_MULTIPLY||Key==VK_DIVIDE||Key==VK_ADD)
	{
		ReleaseKeys();
		DrawKeyboard();
	}
}
//---------------------------------------------------------------------------

void __fastcall TFormMain::SpeedButtonResetClick(TObject *Sender)
{
	bool prev;

	prev=SpeedButtonPause->Down;

	SpeedButtonPause->Down=true;

	if(Application->MessageBox("Reset the system?","Confirm",MB_YESNO)==IDYES)
	{
		aon_init();

		SpeedButtonRun->Down=true;
		SpeedButtonTurbo->Down=false;
	}
	else
	{
		if(!prev) SpeedButtonRun->Down=true;
	}
}
//---------------------------------------------------------------------------

void __fastcall TFormMain::SpeedButtonAdvanceClick(TObject *Sender)
{
	SpeedButtonPause->Down=true;
	SpeedButtonTurbo->Down=false;
}
//---------------------------------------------------------------------------

void __fastcall TFormMain::MNoDiodesClick(TObject *Sender)
{
	AON.diodes=0;

	if(MDiodesVD1->Checked) AON.diodes|=0x80;
	if(MDiodesVD2->Checked) AON.diodes|=0x40;
	if(MDiodesVD3->Checked) AON.diodes|=0x20;
	if(MDiodesVD4->Checked) AON.diodes|=0x10;
	if(MNoDiodes->Checked) AON.diodes|=0x01;

	DrawKeyboard();
}
//---------------------------------------------------------------------------

void __fastcall TFormMain::FormDeactivate(TObject *Sender)
{
	ReleaseKeys();	
}
//---------------------------------------------------------------------------

void __fastcall TFormMain::MViewKeypadClick(TObject *Sender)
{
	UpdateWindowsView();
}
//---------------------------------------------------------------------------

void __fastcall TFormMain::MLoadROMClick(TObject *Sender)
{
	if(!OpenDialogROM->Execute()) return;

	romName=OpenDialogROM->FileName;

	LoadROM();
	UpdateWindowsView();
}
//---------------------------------------------------------------------------

void __fastcall TFormMain::Timer1Timer(TObject *Sender)
{
	if(!UpdateActive) DrawDebugInfo(ImageDebug);
	if(!UpdateActive) ImageDisplay->Repaint();
}
//---------------------------------------------------------------------------

void __fastcall TFormMain::SpeedButtonHandsetDownClick(TObject *Sender)
{
	AON.handset_down=SpeedButtonHandsetDown->Down;
}
//---------------------------------------------------------------------------

